#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function
from __future__ import absolute_import

__author__     = "Lluís Vilanova"
__copyright__  = "Copyright 2008-2014, Lluís Vilanova"
__license__    = "GPL version 3 or later"

__maintainer__ = "Lluís Vilanova"
__email__      = "vilanova@ac.upc.edu"


import sciexp2.version


__version_info__ = sciexp2.version.__version_info__
__version__ = sciexp2.version.__version__
